/*
 * Copyright (C) 2011 by egnite GmbH
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*
 * $Id: df_flasher.c 294 2012-01-03 10:18:06Z harald $
 */

#include <dev/board.h>
#include <dev/spi_at45d.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <errno.h>

#include "df_flasher.h"

static NUTDEVICE *devAt45;

static int At45dConfigDevice(void)
{
    if (devAt45 == NULL) {
        NUTSPINODE *node;
        NUTDEVICE *dev;
        dev = &devSpiAt45d0;
        node = (NUTSPINODE *) dev->dev_icb;
        if (node->node_bus == NULL) {
            NUTSPIBUS *bus;
            bus = &DEV_SPIBUS;
            node->node_bus = bus;
        }
        if ((*node->node_bus->bus_initnode) (node)) {
            return -1;
        }
        NutEventPost(&node->node_bus->bus_mutex);
        if (SpiAt45dInit(dev)) {
            return -1;
        }
        devAt45 = dev;
    }
    return 0;
}

/*
 * Get DataFlash page size.
 */
size_t DataFlashPageSize(void)
{
    int rc = 0;

    rc = At45dConfigDevice();
    if (rc == 0) {
        rc = (size_t) SpiAt45dPageSize(devAt45);
    }
    return rc;
}

/*
 * Transfer the image.
 */
int DataFlashFlasher(char *image, uint32_t start, size_t pgsize)
{
    int fd;
    int got;
    uint32_t page;
    void *wbuf;
    int wcnt;
    void *rbuf;
    int rcnt;

    printf("Image '%s': ", image);
    if ((fd = _open(image, _O_RDONLY | _O_BINARY)) == -1) {
        printf("Not available, result=%d.\n", errno);
        return 0;
    }

    page = start / pgsize;
    wbuf = malloc(pgsize);
    rbuf = malloc(pgsize);
    if (wbuf == NULL || rbuf == NULL) {
        printf("Out of memory. Really?\n");
        return -1;
    }

    printf("Starting at %08lX, page %lu\n", start, page);
    wcnt = 0;
    rcnt = 0;
    for (;; page++) {
        if ((got = _read(fd, wbuf, pgsize)) != 0) {
            if (got < 0) {
                printf("Error %d: Reading image file.\n", errno);
                free(wbuf);
                free(rbuf);
                return -1;
            }
            if (got) {
                rcnt += got;
                if (SpiAt45dPageRead(devAt45, page, rbuf, got) != got) {
                    printf("Error: DataFlash read failed at page %lu.\n", page);
                    free(wbuf);
                    free(rbuf);
                    return -1;
                }
                if (memcmp(wbuf, rbuf, got)) {
                    if (got < pgsize) {
                        memset(wbuf + got, 0xFF, pgsize - got);
                    }
                    if (SpiAt45dPageWrite(devAt45, page, wbuf, pgsize) != pgsize) {
                        printf("Error: DataFlash write failed at page %lu.\n", page);
                        free(wbuf);
                        free(rbuf);
                        return -1;
                    }
                    if (SpiAt45dPageRead(devAt45, page, rbuf, pgsize) != pgsize) {
                        printf("Error: DataFlash read failed at page %lu.\n", page);
                        free(wbuf);
                        free(rbuf);
                        return -1;
                    }
                    if (memcmp(wbuf, rbuf, pgsize)) {
                        printf("Error: DataFlash compare failed at page %lu.\n", page);
                        free(wbuf);
                        free(rbuf);
                        return -1;
                    }
                    wcnt += pgsize;
                }
            }
        }
        if (got < pgsize) {
            break;
        }
    }
    _close(fd);
    free(wbuf);
    free(rbuf);
    printf("%d of %d bytes updated.\n", wcnt, rcnt);

    return 0;
}

